import {Net} from "./Net";
import {DefaultConstructor, Json} from "mvx-utils";
import {Method} from "./http/HttpBuilder";
import {SignNonceStr} from "../Sign";
import { Token } from "../model/Token";

export enum Code {
  Success = 200,
  TokenExpire = 401,
  ElseError = 500,
}

export class CodeError extends Error{
  constructor(err:string|Error, public code:Code = Code.ElseError) {
    super();
    this.name = "CodeError";
    if (typeof err === 'string') {
      this.message = `${err}`
    } else {
      this.message = `${err.message}`;
      this.stack = err.stack
    }
  }
}

interface opt {
  body:(body:object)=>[string, Error|null];
  header:Map<string, string>;
  token:(body:object, tokenStr:string)=>void;
  method:Method;
}

export type HttpOption = (o:opt)=>Error|null
function newHttpOpt(...opts:HttpOption[]):opt {
  let opt:opt = {
    header:new Map<string, string>(),
    body:body => {
      throw new Error(`body encode not implement`)
    },
    method:"GET",
    token:tokenStr => {},
  };

  opts.forEach(f=> {
    f(opt);
  });

  return opt;
}

export function WithHeader(header:Map<string, string>):HttpOption {
  return (o:opt):Error|null => {
    header.forEach((value, key) => {
      o.header.set(key, value);
    });
    return null;
  }
}

export function WithBearerToken():HttpOption {
  return (o:opt):Error|null => {
    o.token = (request:object, tokenStr:string) => {
      o.header.set("Authorization", `Bearer ${tokenStr}`);
    };
    return null;
  }
}

export function WithBodyToken(tokenKey:string = "token"):HttpOption {
  return (o:opt):Error|null => {
    o.token = (body:object, tokenStr:string) => {
      (<Record<string, string>>body)[tokenKey] = tokenStr;
    };
    return null;
  }
}

export function WithoutToken():HttpOption {
  return (o:opt):Error|null => {
    o.token = (body:object, tokenStr:string) => {};
    return null;
  }
}

export function WithReqId():HttpOption {
  return (o:opt):Error|null => {
    o.header.set("X-Req-Id", SignNonceStr());
    return null;
  }
}

export function WithBody(body:(body:object)=>[string, Error|null]):HttpOption {
  return (o:opt):Error|null => {
    o.body = body;
    return null;
  }
}

export function WithFormUrlEncodedBody():HttpOption {
  return (o:opt):Error|null => {
    o.header.set("X-Requested-With", "XMLHttpRequest");
    o.header.set('Content-Type', 'application/x-www-form-urlencoded');

    o.body = (body:object):[string, Error|null] => {
      const [json, err] = new Json().toJson(body);
      if (err) {
        return ["", err];
      }

      const args = JSON.parse(json);
      const ret = Object.keys(args).map((key) => {
        if ((<any>args)[key] == null) {
          return encodeURIComponent(key) + "="
        }

        return encodeURIComponent(key) + "=" + encodeURIComponent((<any>args)[key]);
      }).join("&");
      return [ret, null];
    };

    return null;
  }
}

export function WithJsonBody():HttpOption {
  return (o:opt):Error|null => {
    o.header.set("X-Requested-With", "XMLHttpRequest");
    o.header.set('Content-Type', 'application/json;charset=utf8');
    o.body = (body:object):[string, Error|null] => {
      return new Json().toJson(body);
    };

    return null;
  }
}

export function WithHttpMethod(method:Method):HttpOption {
  return (o:opt):Error|null => {
    o.method = method;
    return null;
  }
}

export async function SendNoToken<T extends object>(uri:string,
                                             request:object,
                                             resType:DefaultConstructor<T>,
                                             net:Net, ...opts:HttpOption[]):Promise<[T, CodeError|null]> {
  const opt = newHttpOpt(...opts);
  opt.token(request, await net.getToken());

  const [body, err0] = opt.body(request);
  if (err0) {
    return [new resType(), new CodeError(err0.message)]
  }
  const [res, err] = await net.getHttpBuilder()
    .setUri(uri)
    .setMethod(opt.method)
    .setHeaders(opt.header)
    .setContent(body).build().send();
  if (err.code!==Code.Success) {
    return [new resType(), err]
  }
  if (res === "") {
    return [new resType(), new CodeError("response is empty")]
  }

  let [resT, err1] = new Json().fromJson(res, resType);
  if (err1) {
    return [new resType(), new CodeError(err1)]
  }

  return [resT, null]
}

export async function Send<T extends object>(uri:string,
                                      request:object,
                                      resType:DefaultConstructor<T>,
                                      net:Net, ...opts:HttpOption[]):Promise<[T, CodeError|null]> {
  if (net.is401()) {
    return [new resType(), new CodeError("has 401", Code.TokenExpire)]
  }

  let token = await net.getToken();
  // token为空，需要处理401，并且不执行真正的网络请求。
  if (token === Token.empty) {
    await net.process401();
    return [new resType(), new CodeError("token is empty", Code.TokenExpire)]
  }

  let [ret, err] = await SendNoToken(uri, request, resType, net, ...opts);
  if (!err || err.code !== Code.TokenExpire) {
    // 取消401
    net.clear401();
    return [ret, err]
  }

  // 返回401时，如果请求的token与现在存储token不是一样的，说明有登录接口修改过，最新
  // 存储的token是否过期，无法判断，所以不能执行真正的401操作
  // 此处不能通过net层is401来判断，还没有开始处理401或者其他接口处理完401的情况下，is401都是false
  // 但是后一种情况却不能执行401；对token而言后一种情况token一定不一样，却可以分辨出两种情况
  if (token !== await net.getToken()) {
    return [new resType(), new CodeError("token is too old", Code.TokenExpire)]
  }

  await net.process401();
  return [new resType(), new CodeError("token is expire", Code.TokenExpire)]
}